import {
  svg1,
  svg2,
  slider,
  dropdown,
  continents,
  gdpScale,
  MIN_LIFE,
  MAX_LIFE,
  lifeScale,
  MIN_YEAR,
  yearScale,
  radiusScale,
  colorScale,
  getCircleRadius,
  getContinentColor,
  lineGenerator,
  factBox,
} from './setup.js';

import {
  updateYear,
  highlightMarks,
  resetMarks,
  popUpTooltip,
  adjustTooltip,
  hideTooltip,
  setFactBox,
} from './events.js';

import { facts } from './annotations.js';

import { processRawData } from './preprocess.js';

//////////////////////////////////////////////////////////////////////////
// Setting up initial global data variables (used for dynamic filtering)
//////////////////////////////////////////////////////////////////////////
var data = new Map();
var countryData = new Array();
var curYear = String(MIN_YEAR);
var curContinent = 'global';
var curCountry = '';
var validCountries = new Array();

//////////////////////////////////////////////////////////////////////////
// Hooking up event handlers for my controls (slider, dropdown, keyboard)
//////////////////////////////////////////////////////////////////////////
slider.on('input', function () {
  curYear = this.value;
  updateYear(data, facts, curYear, curContinent);
  renderData(100);
});

dropdown.on('change', function () {
  curContinent = d3.select(this).property('value');
  renderData(0);
});

d3.select('body').on('keydown', function (event) {
  switch (event.key) {
    // I can use a "switch" to easily trigger the same behavior for different keys
    case 'ArrowRight':
    case 'd':
      curYear = String(Math.min(+curYear + 1, 2008));
      updateYear(data, facts, curYear, curContinent);
      renderData(300);
      break;
    case 'ArrowLeft':
      curYear = String(Math.max(+curYear - 1, 1800));
      updateYear(data, facts, curYear, curContinent);
      renderData(300);
      break;
  }
});

//////////////////////////////////////////////////////////////////////////
// Here's the main part of this file: rendering the marks on my plots
// This will handle rendering all the circles and lines for a given year
//////////////////////////////////////////////////////////////////////////
function renderData(duration) {
  // First, I need to pull out the data objects for all countries (the .values()
  // of my Map), convert it to an Array (using the [...] spread operator), and then
  // filter the Array down to just the actual data points I want to draw
  validCountries = countryData.filter(
    (c) =>
      curContinent === 'global' ||
      c.continent === curContinent,
  );

  // Here's the code for rendering the circles on the scatter plot; since I may be dynamically
  // adding or removing some of the data points, first I'll use selectAll('circle') to grab all
  // of the currently drawn circles, and then let D3's data join semantics handle deleting circles
  // for which I've filtered the corresponding country, add circles for each newly included country,
  // and otherwise just update the attributes for valid countries that I already have circles for
  const circleMarks = svg1
    .selectAll('circle')
    .data(validCountries)
    .join('circle')
    .attr('r', 3)
    .attr('stroke', 'black')
    .attr('stroke-width', 0.5)
    .attr('r', (d) => getCircleRadius(d, curYear))
    .attr('opacity', 0.7)
    .attr('fill', (d) => getContinentColor(d, curCountry))
    .attr('class', (d) => d.continent)
    .attr('id', (d) => d.country);

  // I split out these attributes as a separate update since I want all of the above attributes to
  // get set instantly, but I want any updates to the circles' positions to animate over time
  circleMarks
    .transition()
    .duration(duration)
    .attr('cx', (d) => gdpScale(d.income.get(curYear)))
    .attr('cy', (d) => lifeScale(d.life_exp.get(curYear)));

  svg1.on('click', () => {
    curCountry = '';
    d3.select('#pinnedCountry').text('');
    renderData(0);
  });

  // Finally, I'll hook up all the event listeners for my scatter points. The implementation of these
  // methods has been factored out into the "events.js" file for clarity. Look how easy it is to read
  // what all the different mouse events will do to the plot!!
  circleMarks
    .on('mouseenter', function (event, data) {
      highlightMarks(this, data, curYear);
      popUpTooltip(event, data, curYear);
    })
    .on('mousemove', function (event, data) {
      adjustTooltip(event);
    })
    .on('mouseleave', function (event, data) {
      resetMarks(this, data, curYear, curCountry);
      hideTooltip();
    })
    .on('click', (event, data) => {
      event.stopPropagation();
      curCountry = data.country;
      curContinent = data.continent;
      dropdown.property('value', curContinent);
      d3.select('#pinnedCountry').text(
        `Selected: ${curCountry}`,
      );
      renderData(0);
    });

  // All that's left to do is draw the lines; again D3's .data(...).join(...) convention will
  // automatically delete lines when I filter out their country data, and append them whenever
  // the data is filtered back in
  const lineMarks = svg2
    .selectAll('path.line')
    .data(validCountries)
    .join('path')
    .attr('opacity', 0.1)
    .attr('fill', 'none')
    .attr('stroke-width', 2)
    .attr('stroke', (d) => getContinentColor(d, curCountry))
    // I want these lines to have two classes: one so that it's easy to select all the lines for
    // a given continent, and another class that I can use in selections to avoid accidentally selecting
    // the lines in the axis/margins when all I want is the data lines
    .attr('class', (d) => d.continent)
    .classed('line', true)
    .attr('id', (d) => d.country)
    .attr('d', (d) => lineGenerator(d.life_exp.entries()));

  annotatePlot();
}

function annotatePlot() {
  drawYearLine();
  drawGlobalAverageDot();
  drawContinentAverages();
}

function drawYearLine() {
  svg2
    .selectAll('#yearMarkerLine')
    .data([
      [
        // This is a line with two (x,y) coordinates that is vertical at the currentYear position
        [curYear, MIN_LIFE],
        [curYear, MAX_LIFE],
      ],
    ])
    .join('path')
    .attr('opacity', 0.8)
    .attr('fill', 'none')
    .attr('stroke-width', 1.5)
    .attr('stroke', 'black')
    .attr('id', 'yearMarkerLine')
    .attr('d', (d) => lineGenerator(d));
}

function drawGlobalAverageDot() {
  // And then lets draw a circle for the global average this year as well
  let globalAverage = d3.mean(
    d3.map(validCountries, (c) => c.life_exp.get(curYear)),
  );
  svg2
    .selectAll('#globalAverageMark')
    .data([null])
    .join('circle')
    .attr('r', 5)
    .attr('fill', 'none')
    .attr('stroke', 'black')
    .attr('stroke-width', 2)
    .attr('cx', yearScale(curYear))
    .attr('cy', lifeScale(globalAverage))
    .attr('id', 'globalAverageMark');
}

//Read the data
const countryStatsUrl =
  'https://raw.githubusercontent.com/adamjanes/udemy-d3/master/05/5.10.1/data/data.json';
d3.json(countryStatsUrl).then((rawData) => {
  data = processRawData(rawData);
  let avgGDP = (c) => d3.mean(c.income.values());
  let avgLife = (c) => d3.mean(c.life_exp.values());
  let filterFunc = (c) =>
    avgGDP(c) > 100 &&
    avgLife(c) > 20 &&
    c.country != 'Croatia';
  let sortFunc = (c) => +c.population.get('1990');
  countryData = d3
    .sort([...data.values()].filter(filterFunc), sortFunc)
    .reverse();
  console.log(data);
  renderData(0);
});
